/**
 * This file contains a backup system for finding necessary trial information if
 * your matching service does not provide it.
 *
 * The intended usages is essentially:
 *
 * Create the service:
 *
 * const ctgService = await createClinicalTrialsGovService('temp-data');
 *
 * Then, when ResearchStudy objects exist that need to be filled out, assuming
 * that studies is a ResearchStudy[]:
 *
 * await ctgService.updateResearchStudies(studies);
 *
 * This will fill out whatever can be filled out within the given studies.
 */
import { ResearchStudy } from 'fhir/r4';
import { SearchBundleEntry as SearchSetEntry } from './searchset';
import { ClinicalTrialsGovAPI, Study } from './clinicaltrialsgov-api';
import * as sqlite from 'sqlite';
/**
 * Logger type from the NodeJS utilities. (The TypeScript definitions for Node
 * don't bother naming this type.)
 */
type Logger = (message: string, ...param: unknown[]) => void;
/**
 * More for documentation purposes than anything else, this marks strings that are expected to be a valid NCT number.
 * (Generally they aren't checked to ensure they're valid, the methods involved will return nothing if they are not.)
 */
export type NCTNumber = string;
/**
 * The URL for the "system" field of identifiers that indicate the
 * ClinicalTrial.gov identifier, or NTC number, of a clinical trial.
 */
export declare const CLINICAL_TRIAL_IDENTIFIER_CODING_SYSTEM_URL = "http://clinicaltrials.gov/";
/**
 * Checks to ensure that a given NCT number is valid. This checks to make sure
 * it matches the regular expression /^NCT[0-9]{8}$/.
 * @param nctNumber the NCT number to check
 */
export declare function isValidNCTNumber(nctNumber: string): boolean;
export declare function parseNCTNumber(nctNumber: string): number | undefined;
export declare function formatNCTNumber(nctNumber: number): string;
/**
 * Finds the NCT number specified for the given ResearchStudy, assuming there is
 * one. This requires an identifier on the given ResearchStudy that either
 * belongs to the coding system "http://clinicaltrials.gov/" or is a valid NCT
 * number as determined by #isValidNCTNumber.
 *
 * This will return the value of the first identifier belonging to the coding
 * system "http://clinicaltrials.gov/", and if there is no identifier, return
 * the first identifier value that is considered valid.
 *
 * @param study the research study
 * @returns the NCT number or null if none
 */
export declare function findNCTNumber(study: ResearchStudy): NCTNumber | null;
/**
 * Finds all the NCT numbers within the given list of studies. Returns a map of
 * NCT numbers to the research studies they match.
 * @param studies the NCT numbers found, if any
 */
export declare function findNCTNumbers(studies: ResearchStudy[]): Set<NCTNumber>;
/**
 * Attempts to parse the given string as a Study JSON object.
 * @param fileContents the contents of the file as a string
 * @param log an optional logger
 * @returns a promise that resolves to the parsed object or null if the study
 * could not be parsed
 */
export declare function parseStudyJson(fileContents: string, log?: Logger): Study | null;
export interface ClinicalTrialsGovServiceOptions {
    /**
     * The logging function to use to log debug information. If not set the default is `debuglog('ctgovservice')` from
     * the `util` module in Node.js.
     */
    log?: Logger;
    /**
     * Time in milliseconds to expire files after.
     */
    expireAfter?: number;
    /**
     * Interval in milliseconds to run periodic cache clean-ups. If set to 0 or infinity, never run cleanups.
     */
    cleanInterval?: number;
    /**
     * If given, rather than use the default Node.js file system, use the given override. Mainly used for tests but may
     * also be useful in instances where a "real" file system is unavailable.
     */
    fs?: FileSystem;
}
/**
 * System to fill in data on research studies based on the trial data from
 * https://clinicaltrials.gov/
 *
 * The current system works as follows:
 *
 * Call updateResearchStudies() with the ResearchStudy objects to update. It will locate all ResearchStudy objects that
 * contain an NCT ID. It will then:
 *
 * 1. Call downloadTrials() to download the ClinicalStudy data from clinicaltrials.gov. It starts an HTTPS request to
 *    download the trial results and, if successful, passes the response stream off to extractResults().
 * 2. extractResults() saves the stream its given to a temporary ZIP file and extracts that file to a temporary
 *    directory. Once the ZIP is extracted, it deletes the temporary ZIP. It then returns the directory where the files
 *    were extracted.
 * 3. With the path to the extracted ZIP, updateResearchStudies() can now use the downloaded trial data to update the
 *    trials.
 * 4. Before resolving, the temporary directory is deleted.
 *
 * A future version will likely change this to make it so that rather than immediately deleting the temporary trial
 * data, it's kept around in a least-recently-used cache and only deleted as the cache exceeds a given size. This will
 * add a few new public APIs to enable the retrieval of clinical study data by NCT ID. Right now, though, there is no
 * way to extract the clinical study data directly.
 */
export declare class ClinicalTrialsGovService {
    readonly db: string | sqlite.Database;
    /**
     * Log used to log debug information. Either passed in at creation time or created via the Node.js util.debuglog
     * function. With the latter, activate by setting the NODE_DEBUG environment variable to "ctgovservice" (or include it
     * in a comma separated list of debugging modules to include).
     */
    readonly log: Logger;
    private _maxTrialsPerRequest;
    /**
     * The maximum number of trials to attempt to download in a single request. Each NCT ID increases the URI size by
     * effectively 12 bytes: essentially each NCT ID becomes ',NCT12345678'. The request path "counts" as a header and
     * has to be smaller than maxHeaderSize given to the client. By default, this uses the default client, so the client
     * should be the default HTTPS client. In theory this would support around 500 per request, instead this defaults to
     * 128 which is a nice round number. (In binary, at least.)
     */
    get maxTrialsPerRequest(): number;
    set maxTrialsPerRequest(newValue: number);
    private _expirationTimeout;
    /**
     * The expiration timeout, the time in milliseconds after a cache entry should be considered expired. The default
     * is 1 hour.
     */
    get expirationTimeout(): number;
    set expirationTimeout(value: number);
    private _cleanupIntervalMillis;
    /**
     * The interval between cleanup sweeps. Set to 0 (or any negative value) or Infinity to disable periodic cleanup. Note
     * that the value is limited to the range 60000 (a minute)-2147483647 (the maximum value of a signed 32-bit integer)
     * as that maximum value is the maximum allowed timeout for a timer within Node.js. If the value is set to a negative
     * number or Infinity, the retrieved value will be 0, to indicate that the interval will not be set.
     */
    get cleanupInterval(): number;
    set cleanupInterval(value: number);
    /**
     * The maximum allowed entry size before it is rejected. Default is 128MB, which is hopefully well more than necessary
     * for any reasonable clinical trial JSON.
     */
    maxAllowedEntrySize: number;
    private cacheDBPath;
    private cacheDB;
    /**
     * Internal flag indicating if init() has been called
     */
    private cacheReady;
    private cleanupTimeout;
    readonly service: ClinicalTrialsGovAPI;
    /**
     * Creates a new instance. This will not initialize the cache or load anything, this merely creates the object. Use
     * the #init() method to initialize the service and load existing data. (Or use createClinicalTrialsGovService() do
     * construct and initialize at the same time.)
     *
     * The {@link #init} method must be used **even if** given a database object, as it will initialize the necessary
     * tables within the database. If given a path name, calling {@link #destroy} will close the database. However, if
     * given a database directly, {@link #destroy} **will not** close the database and it will be the responsibility of
     * the code using the service to close it.
     *
     * @param db either the path to the database file to use, or the sqlite.Database to use directly.
     * @param log an optional function to receive debug log messages. By default this uses util.debuglog('ctgovservice')
     *     meaning that the log can be activated by setting NODE_DEBUG to "ctgovservice"
     */
    constructor(db: string | sqlite.Database, options?: ClinicalTrialsGovServiceOptions);
    /**
     * Initializes the service. This will open the SQLite database and run through any necessary data migrations for it.
     */
    init(): Promise<void>;
    /**
     * Migrate a database. The arguments exist mostly for typing/testing purposes.
     * @param db the database to migrate
     * @param migrations the migrations to run (they're run in key order). This is an argument for testing purposes.
     */
    private migrateDB;
    /**
     * Internal method to get the database object, throwing an exception if it's
     * not available.
     * @returns the database
     */
    private getDB;
    /**
     * Sets the timeout interval assuming it was set.
     */
    private setCleanupTimeout;
    /**
     * Shuts down the service, doing any final necessary cleanup. If given a database object when constructed, this
     * **will not** close the database. However, future calls to this object will no longer function.
     */
    destroy(): Promise<void>;
    /**
     * Attempts to update the given research studies with data from within this store. This returns a Promise that
     * resolves when the research studies have been updated with any available additional data. The Promise may be
     * rejected if the process fails entirely (that is, clinicaltrials.gov could not be contacted at all, or an I/O error
     * occurs).
     *
     * @param studies the studies to attempt to update
     * @returns a Promise that resolves with updated studies
     */
    updateResearchStudies(studies: ResearchStudy[]): Promise<ResearchStudy[]>;
    updateSearchSetEntries(entries: SearchSetEntry[]): Promise<SearchSetEntry[]>;
    /**
     * Tells the cache to delete all expired cached files. Currently this does nothing - entries never expire. It may
     * make sense to clean up the database every once and a while, but for now, this is a no-op.
     */
    removeExpiredCacheEntries(): Promise<void>;
    /**
     * Returns a list of NCT IDs that are valid and that are not in the cache.
     * @param ids the NCT IDs to locate in the cache
     * @return a list containing only NCT IDs that are valid and missing
     */
    findCacheMisses(ids: string[]): Promise<string[]>;
    /**
     * Ensures that a given set of NCT IDs is available within the service, assuming they exist on ClinicalTrials.gov.
     * @param ids the IDs to ensure are available
     * @returns a Promise that resolves once any downloads have completed
     */
    ensureTrialsAvailable(ids: string[]): Promise<void>;
    ensureTrialsAvailable(studies: ResearchStudy[]): Promise<void>;
    /**
     * A Promise used as a lock to ensure that only one downloadTrials is attempting to insert cache entries at a time.
     */
    private _downloadTrialsLock;
    /**
     * Downloads the given trials from ClinicalTrials.gov and stores them in the data directory. Note that this will
     * always replace trials if they exist.
     *
     * @param ids the IDs of the trials to download
     * @returns true if the request succeeded, false if it failed for some reason - the exact reason can be logged but
     * is otherwise silently eaten
     */
    protected downloadTrials(ids: string[]): Promise<boolean>;
    private tryFetchStudies;
    private addCacheEntry;
    /**
     * Loads a ClinicalStudy from an extracted dataset. This will never download a copy, this will only ever return from
     * within the cache.
     * @param nctNumber the NCT number
     * @returns a Promise that resolves to either the parsed ClinicalStudy or to null if the ClinicalStudy does not exist
     */
    getCachedClinicalStudy(nctNumber: NCTNumber): Promise<Study | null>;
    /**
     * Creates and initializes a new service for retrieving data from http://clinicaltrials.gov/.
     *
     * This is essentially the same as constructing the object and then calling init() on it.
     *
     * Note the same behavior applies when `create()` is invoked with a database: the database object is nulled out when
     * {@link #destroy} is called, but it isn't closed. When invoked with a database object, it is the caller's
     * responsibility to close it.
     *
     * @param db the path to the database file or the database to use directly
     * @param options additional options that can be set to further configure the trial service
     * @returns a Promise that resolves when the service is ready
     */
    static create(db: string | sqlite.Database, options?: ClinicalTrialsGovServiceOptions): Promise<ClinicalTrialsGovService>;
}
/**
 * Alias of ClinicalTrialsGovService.create.
 */
export declare const createClinicalTrialsGovService: typeof ClinicalTrialsGovService.create;
export {};
