"use strict";
/**
 * This file contains a backup system for finding necessary trial information if
 * your matching service does not provide it.
 *
 * The intended usages is essentially:
 *
 * Create the service:
 *
 * const ctgService = await createClinicalTrialsGovService('temp-data');
 *
 * Then, when ResearchStudy objects exist that need to be filled out, assuming
 * that studies is a ResearchStudy[]:
 *
 * await ctgService.updateResearchStudies(studies);
 *
 * This will fill out whatever can be filled out within the given studies.
 */
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.createClinicalTrialsGovService = exports.ClinicalTrialsGovService = exports.parseStudyJson = exports.findNCTNumbers = exports.findNCTNumber = exports.formatNCTNumber = exports.parseNCTNumber = exports.isValidNCTNumber = exports.CLINICAL_TRIAL_IDENTIFIER_CODING_SYSTEM_URL = void 0;
const util_1 = require("util");
const clinicaltrialsgov_api_1 = require("./clinicaltrialsgov-api");
const study_fhir_converter_1 = require("./study-fhir-converter");
const sqlite = require("sqlite");
const sqlite3 = require("sqlite3");
/**
 * The URL for the "system" field of identifiers that indicate the
 * ClinicalTrial.gov identifier, or NTC number, of a clinical trial.
 */
exports.CLINICAL_TRIAL_IDENTIFIER_CODING_SYSTEM_URL = 'http://clinicaltrials.gov/';
/**
 * Checks to ensure that a given NCT number is valid. This checks to make sure
 * it matches the regular expression /^NCT[0-9]{8}$/.
 * @param nctNumber the NCT number to check
 */
function isValidNCTNumber(nctNumber) {
    return /^NCT[0-9]{8}$/.test(nctNumber);
}
exports.isValidNCTNumber = isValidNCTNumber;
function parseNCTNumber(nctNumber) {
    if (isValidNCTNumber(nctNumber)) {
        return parseInt(nctNumber.substring(3));
    }
    else {
        return undefined;
    }
}
exports.parseNCTNumber = parseNCTNumber;
function formatNCTNumber(nctNumber) {
    // Make sure the NCT number is an integer
    nctNumber = Math.floor(nctNumber);
    if (nctNumber < 0 || nctNumber > 99999999) {
        throw new Error(`Invalid NCT number ${nctNumber}`);
    }
    return `NCT${nctNumber.toFixed(0).padStart(8, '0')}`;
}
exports.formatNCTNumber = formatNCTNumber;
/**
 * Finds the NCT number specified for the given ResearchStudy, assuming there is
 * one. This requires an identifier on the given ResearchStudy that either
 * belongs to the coding system "http://clinicaltrials.gov/" or is a valid NCT
 * number as determined by #isValidNCTNumber.
 *
 * This will return the value of the first identifier belonging to the coding
 * system "http://clinicaltrials.gov/", and if there is no identifier, return
 * the first identifier value that is considered valid.
 *
 * @param study the research study
 * @returns the NCT number or null if none
 */
function findNCTNumber(study) {
    if (study.identifier && Array.isArray(study.identifier) && study.identifier.length > 0) {
        for (const identifier of study.identifier) {
            if (identifier.system === exports.CLINICAL_TRIAL_IDENTIFIER_CODING_SYSTEM_URL &&
                typeof identifier.value === 'string' &&
                isValidNCTNumber(identifier.value))
                return identifier.value;
        }
        // Fallback: regexp match
        for (const identifier of study.identifier) {
            if (typeof identifier.value === 'string' && isValidNCTNumber(identifier.value))
                return identifier.value;
        }
    }
    // Return null on failures
    return null;
}
exports.findNCTNumber = findNCTNumber;
/**
 * Finds all the NCT numbers within the given list of studies. Returns a map of
 * NCT numbers to the research studies they match.
 * @param studies the NCT numbers found, if any
 */
function findNCTNumbers(studies) {
    const result = new Set();
    for (const study of studies) {
        const nctId = findNCTNumber(study);
        if (nctId !== null) {
            result.add(nctId);
        }
    }
    return result;
}
exports.findNCTNumbers = findNCTNumbers;
/**
 * Attempts to parse the given string as a Study JSON object.
 * @param fileContents the contents of the file as a string
 * @param log an optional logger
 * @returns a promise that resolves to the parsed object or null if the study
 * could not be parsed
 */
function parseStudyJson(fileContents, log) {
    try {
        const json = JSON.parse(fileContents);
        if (typeof json === 'object' && !Array.isArray(json)) {
            // "null" is currently "valid" - a cached failure?
            return json;
        }
        else {
            if (log) {
                log('Invalid JSON object for Study: %o', json);
            }
            return null;
        }
    }
    catch (ex) {
        if (log) {
            log('Unable to parse JSON object: %o', ex);
        }
        return null;
    }
}
exports.parseStudyJson = parseStudyJson;
// Order matters in this
const MIGRATIONS = {
    init: {
        up: (db) => __awaiter(void 0, void 0, void 0, function* () {
            // Create the cache table
            yield db.run(`CREATE TABLE ctgov_studies
        (
          nct_id INTEGER PRIMARY KEY,
          study_json TEXT,
          error_message TEXT,
          created_at NUMERIC NOT NULL
        )`);
        })
    }
};
/**
 * System to fill in data on research studies based on the trial data from
 * https://clinicaltrials.gov/
 *
 * The current system works as follows:
 *
 * Call updateResearchStudies() with the ResearchStudy objects to update. It will locate all ResearchStudy objects that
 * contain an NCT ID. It will then:
 *
 * 1. Call downloadTrials() to download the ClinicalStudy data from clinicaltrials.gov. It starts an HTTPS request to
 *    download the trial results and, if successful, passes the response stream off to extractResults().
 * 2. extractResults() saves the stream its given to a temporary ZIP file and extracts that file to a temporary
 *    directory. Once the ZIP is extracted, it deletes the temporary ZIP. It then returns the directory where the files
 *    were extracted.
 * 3. With the path to the extracted ZIP, updateResearchStudies() can now use the downloaded trial data to update the
 *    trials.
 * 4. Before resolving, the temporary directory is deleted.
 *
 * A future version will likely change this to make it so that rather than immediately deleting the temporary trial
 * data, it's kept around in a least-recently-used cache and only deleted as the cache exceeds a given size. This will
 * add a few new public APIs to enable the retrieval of clinical study data by NCT ID. Right now, though, there is no
 * way to extract the clinical study data directly.
 */
class ClinicalTrialsGovService {
    /**
     * The maximum number of trials to attempt to download in a single request. Each NCT ID increases the URI size by
     * effectively 12 bytes: essentially each NCT ID becomes ',NCT12345678'. The request path "counts" as a header and
     * has to be smaller than maxHeaderSize given to the client. By default, this uses the default client, so the client
     * should be the default HTTPS client. In theory this would support around 500 per request, instead this defaults to
     * 128 which is a nice round number. (In binary, at least.)
     */
    get maxTrialsPerRequest() {
        return this._maxTrialsPerRequest;
    }
    set maxTrialsPerRequest(newValue) {
        // Basically, just make sure the new value is positive and an integer. Positive infinity is fine and remains
        // positive infinity: it means never split.
        if (newValue > 1) {
            this._maxTrialsPerRequest = Math.floor(newValue);
        }
    }
    /**
     * The expiration timeout, the time in milliseconds after a cache entry should be considered expired. The default
     * is 1 hour.
     */
    get expirationTimeout() {
        return this._expirationTimeout;
    }
    set expirationTimeout(value) {
        // Ensure it's at least 1000ms
        this._expirationTimeout = Math.max(value, 1000);
    }
    /**
     * The interval between cleanup sweeps. Set to 0 (or any negative value) or Infinity to disable periodic cleanup. Note
     * that the value is limited to the range 60000 (a minute)-2147483647 (the maximum value of a signed 32-bit integer)
     * as that maximum value is the maximum allowed timeout for a timer within Node.js. If the value is set to a negative
     * number or Infinity, the retrieved value will be 0, to indicate that the interval will not be set.
     */
    get cleanupInterval() {
        return this._cleanupIntervalMillis;
    }
    set cleanupInterval(value) {
        // Clamp it to 0 (special value for "never") or at least 60000 - also clamp the maximum to a 32-bit signed integer
        // which is the maximum allowed timeout within Node.js. (Exceeding that value changes the timeout to 1ms. Surprise!
        // I mean, sure, that's documented, but...)
        this._cleanupIntervalMillis =
            value <= 0 ? 0 : value === Infinity ? 0 : Math.min(Math.max(Math.floor(value), 60000), 0x7fffffff);
    }
    /**
     * Creates a new instance. This will not initialize the cache or load anything, this merely creates the object. Use
     * the #init() method to initialize the service and load existing data. (Or use createClinicalTrialsGovService() do
     * construct and initialize at the same time.)
     *
     * The {@link #init} method must be used **even if** given a database object, as it will initialize the necessary
     * tables within the database. If given a path name, calling {@link #destroy} will close the database. However, if
     * given a database directly, {@link #destroy} **will not** close the database and it will be the responsibility of
     * the code using the service to close it.
     *
     * @param db either the path to the database file to use, or the sqlite.Database to use directly.
     * @param log an optional function to receive debug log messages. By default this uses util.debuglog('ctgovservice')
     *     meaning that the log can be activated by setting NODE_DEBUG to "ctgovservice"
     */
    constructor(db, options) {
        var _a, _b;
        this.db = db;
        this._maxTrialsPerRequest = 128;
        this._expirationTimeout = 60 * 60 * 1000;
        this._cleanupIntervalMillis = 60 * 60 * 1000;
        /**
         * The maximum allowed entry size before it is rejected. Default is 128MB, which is hopefully well more than necessary
         * for any reasonable clinical trial JSON.
         */
        this.maxAllowedEntrySize = 128 * 1024 * 1024;
        /**
         * Internal flag indicating if init() has been called
         */
        this.cacheReady = false;
        this.cleanupTimeout = null;
        /**
         * A Promise used as a lock to ensure that only one downloadTrials is attempting to insert cache entries at a time.
         */
        this._downloadTrialsLock = null;
        if (typeof db === 'string') {
            this.cacheDBPath = db;
            this.cacheDB = null;
        }
        else {
            this.cacheDBPath = null;
            this.cacheDB = db;
        }
        const log = options ? options.log : undefined;
        // If no log was given, create it
        this.log = log !== null && log !== void 0 ? log : (0, util_1.debuglog)('ctgovservice');
        // Default expiration timeout to an hour
        this.expirationTimeout = (_a = options === null || options === void 0 ? void 0 : options.expireAfter) !== null && _a !== void 0 ? _a : 60 * 60 * 1000;
        // Default cleanup interval to an hour
        this.cleanupInterval = (_b = options === null || options === void 0 ? void 0 : options.cleanInterval) !== null && _b !== void 0 ? _b : 60 * 60 * 1000;
        this.service = new clinicaltrialsgov_api_1.ClinicalTrialsGovAPI({ logger: this.log });
    }
    /**
     * Initializes the service. This will open the SQLite database and run through any necessary data migrations for it.
     */
    init() {
        return __awaiter(this, void 0, void 0, function* () {
            if (this.cacheReady) {
                throw new Error('init() has already been called');
            }
            // Technically this isn't really ready yet, but prevent double-calls to init
            this.cacheReady = true;
            let db;
            if (this.cacheDBPath === null) {
                this.log('Using existing database object for clinicaltrials.gov data');
                if (this.cacheDB === null) {
                    throw new Error('Invalid internal state: both cache DB path and DB are null');
                }
                db = this.cacheDB;
            }
            else {
                this.log('Using %s as cache DB for clinicaltrials.gov data', this.cacheDBPath);
                this.cacheDB = db = yield sqlite.open({ driver: sqlite3.Database, filename: this.cacheDBPath });
            }
            yield this.migrateDB(db, MIGRATIONS);
            // Once started, run the cache cleanup every _cleanupIntervalMillis
            this.setCleanupTimeout();
        });
    }
    /**
     * Migrate a database. The arguments exist mostly for typing/testing purposes.
     * @param db the database to migrate
     * @param migrations the migrations to run (they're run in key order). This is an argument for testing purposes.
     */
    migrateDB(db, migrations) {
        return __awaiter(this, void 0, void 0, function* () {
            this.log('Applying migrations...');
            // lock the database
            yield db.run('BEGIN TRANSACTION');
            try {
                yield db.run('CREATE TABLE IF NOT EXISTS migrations (id INTEGER PRIMARY KEY, name TEXT NOT NULL)');
                const appliedMigration = new Set((yield db.all('SELECT name FROM migrations ORDER BY id ASC')).map((row) => row.name));
                for (const name in migrations) {
                    if (appliedMigration.has(name)) {
                        this.log('Migration %s already applied', name);
                    }
                    else {
                        this.log('Applying migration %s...', name);
                        yield migrations[name].up(db);
                        // Insert that this migration was run
                        yield db.run('INSERT INTO migrations (name) VALUES (?)', name);
                        this.log('Migration %s completed.', name);
                    }
                }
                yield db.run('COMMIT');
                this.log('All migrations applied successfully.');
            }
            catch (ex) {
                this.log('Exception applying migrations: %o', ex);
                // Rollback if we can
                yield db.run('ROLLBACK');
                this.log('Migrations were rolled back.');
                throw ex;
            }
        });
    }
    /**
     * Internal method to get the database object, throwing an exception if it's
     * not available.
     * @returns the database
     */
    getDB() {
        if (this.cacheDB === null) {
            throw new Error(`Database not available (${this.cacheReady ? 'destroy() has been called' : 'init() has not been called'})`);
        }
        return this.cacheDB;
    }
    /**
     * Sets the timeout interval assuming it was set.
     */
    setCleanupTimeout() {
        // If set to infinity we internally set it to 0
        if (this._cleanupIntervalMillis > 0) {
            this.cleanupTimeout = setTimeout(() => {
                this.removeExpiredCacheEntries().then(() => {
                    // Set up to do this again when that's done
                    this.setCleanupTimeout();
                }, (error) => {
                    this.log('Error cleaning expired cache entries: %o', error);
                    // Even though this attempt failed, try again later
                    this.setCleanupTimeout();
                });
            }, this._cleanupIntervalMillis);
        }
    }
    /**
     * Shuts down the service, doing any final necessary cleanup. If given a database object when constructed, this
     * **will not** close the database. However, future calls to this object will no longer function.
     */
    destroy() {
        return __awaiter(this, void 0, void 0, function* () {
            if (this.cleanupTimeout !== null) {
                clearTimeout(this.cleanupTimeout);
                // And blank it
                this.cleanupTimeout = null;
            }
            if (this.cacheDBPath === null) {
                // Always null out the database
                this.cacheDB = null;
            }
            else {
                const db = this.cacheDB;
                if (db) {
                    this.cacheDB = null;
                    yield db.close();
                }
            }
        });
    }
    /**
     * Attempts to update the given research studies with data from within this store. This returns a Promise that
     * resolves when the research studies have been updated with any available additional data. The Promise may be
     * rejected if the process fails entirely (that is, clinicaltrials.gov could not be contacted at all, or an I/O error
     * occurs).
     *
     * @param studies the studies to attempt to update
     * @returns a Promise that resolves with updated studies
     */
    updateResearchStudies(studies) {
        return __awaiter(this, void 0, void 0, function* () {
            const nctIds = Array.from(findNCTNumbers(studies));
            if (nctIds.length === 0) {
                // Nothing to do
                return studies;
            }
            this.log('Updating research studies with NCT IDs %j', nctIds);
            // Make sure the NCT numbers are in the cache
            yield this.ensureTrialsAvailable(nctIds);
            // Update the items in the list
            const updatedTrials = new Map();
            yield Promise.all(Array.from(nctIds).map((nctId) => __awaiter(this, void 0, void 0, function* () {
                const study = yield this.getCachedClinicalStudy(nctId);
                if (study) {
                    updatedTrials.set(nctId, (0, study_fhir_converter_1.createResearchStudyFromClinicalStudy)(study));
                }
            })));
            return studies.map((study) => {
                const id = findNCTNumber(study);
                if (id) {
                    // It's possible some IDs may not exist and may be left alone
                    const updated = updatedTrials.get(id);
                    if (updated) {
                        return updated;
                    }
                }
                return study;
            });
        });
    }
    updateSearchSetEntries(entries) {
        return __awaiter(this, void 0, void 0, function* () {
            const studies = entries.map((item) => item.resource);
            yield this.ensureTrialsAvailable(studies);
            yield Promise.all(entries.map((entry) => __awaiter(this, void 0, void 0, function* () {
                const resource = entry.resource;
                const nctId = findNCTNumber(resource);
                if (nctId) {
                    const study = yield this.getCachedClinicalStudy(nctId);
                    if (study) {
                        entry.resource = (0, study_fhir_converter_1.createResearchStudyFromClinicalStudy)(study);
                    }
                }
            })));
            return entries;
        });
    }
    /**
     * Tells the cache to delete all expired cached files. Currently this does nothing - entries never expire. It may
     * make sense to clean up the database every once and a while, but for now, this is a no-op.
     */
    removeExpiredCacheEntries() {
        return __awaiter(this, void 0, void 0, function* () {
            // For now, does nothing.
        });
    }
    /**
     * Returns a list of NCT IDs that are valid and that are not in the cache.
     * @param ids the NCT IDs to locate in the cache
     * @return a list containing only NCT IDs that are valid and missing
     */
    findCacheMisses(ids) {
        return __awaiter(this, void 0, void 0, function* () {
            const db = this.getDB();
            const checkIds = [];
            const idSet = new Set();
            let sql = '';
            for (const id of ids) {
                const nctNum = parseNCTNumber(id);
                if (nctNum !== undefined) {
                    checkIds.push(nctNum);
                    idSet.add(id);
                    if (sql.length === 0) {
                        sql = 'SELECT nct_id FROM ctgov_studies WHERE nct_id IN (?';
                    }
                    else {
                        sql += ', ?';
                    }
                }
            }
            // No valid IDs
            if (sql.length === 0) {
                return [];
            }
            const cacheHitIds = yield db.all(sql + ')', checkIds);
            // Remove each hit ID
            for (const row of cacheHitIds) {
                idSet.delete(formatNCTNumber(row.nct_id));
            }
            return Array.from(idSet);
        });
    }
    ensureTrialsAvailable(idsOrStudies) {
        return __awaiter(this, void 0, void 0, function* () {
            // We only want string IDs and we may end up filtering some of them out
            let ids = [];
            for (const o of idsOrStudies) {
                if (typeof o === 'string') {
                    if (isValidNCTNumber(o))
                        ids.push(o);
                }
                else {
                    const id = findNCTNumber(o);
                    if (id)
                        ids.push(id);
                }
            }
            // Find the IDs we need to fetch
            ids = yield this.findCacheMisses(ids);
            // Now that we have the IDs, we can split them into download requests
            // Run the download requests in series
            for (let start = 0; start < ids.length; start += this.maxTrialsPerRequest) {
                yield this.downloadTrials(ids.slice(start, Math.min(start + this.maxTrialsPerRequest, ids.length)));
            }
        });
    }
    /**
     * Downloads the given trials from ClinicalTrials.gov and stores them in the data directory. Note that this will
     * always replace trials if they exist.
     *
     * @param ids the IDs of the trials to download
     * @returns true if the request succeeded, false if it failed for some reason - the exact reason can be logged but
     * is otherwise silently eaten
     */
    downloadTrials(ids) {
        return __awaiter(this, void 0, void 0, function* () {
            this.log('Fetching studies with NCT IDs %j', ids);
            const studies = yield this.tryFetchStudies(ids);
            // If the call failed, return false
            if (studies === null) {
                return false;
            }
            const db = this.getDB();
            // Some promise magic - if the lock exists, await it
            while (this._downloadTrialsLock !== null) {
                this.log('Waiting on transaction lock for NCT IDs %j', ids);
                yield this._downloadTrialsLock;
            }
            // Now, create the lock. (Default value gives it a type. Promise always runs the function in the constructor
            // immediately so resolveLock will be set to the resolve method, but TypeScript can't prove that.)
            let resolveLock = /* istanbul ignore next */ () => { };
            this._downloadTrialsLock = new Promise((resolve) => {
                resolveLock = resolve;
            });
            try {
                // Surround this in a transaction (otherwise each insert would be a transaction on its own)
                yield db.run('BEGIN');
                try {
                    for (const study of studies) {
                        yield this.addCacheEntry(db, study);
                    }
                    yield db.run('COMMIT');
                    return true;
                }
                catch (ex) {
                    this.log('Exception while adding cache entries: %j', ex);
                    yield db.run('ROLLBACK');
                    return false;
                }
            }
            finally {
                this._downloadTrialsLock = null;
                resolveLock();
            }
        });
    }
    tryFetchStudies(ids) {
        return __awaiter(this, void 0, void 0, function* () {
            try {
                return yield this.service.fetchStudies(ids, this._maxTrialsPerRequest);
            }
            catch (ex) {
                this.log('Error fetching trials from server: %o', ex);
                return null;
            }
        });
    }
    addCacheEntry(db, study) {
        var _a, _b;
        return __awaiter(this, void 0, void 0, function* () {
            // See if we can locate an NCT number for this study.
            const nctNumber = (_b = (_a = study.protocolSection) === null || _a === void 0 ? void 0 : _a.identificationModule) === null || _b === void 0 ? void 0 : _b.nctId;
            if (typeof nctNumber !== 'string') {
                this.log('Ignoring study object from server: unable to locate an NCT ID for it! (protocolSection.identificationModule.nctId missing or not a string)');
                return;
            }
            // Convert to the numeric ID sqlite will actually use
            const rowId = parseNCTNumber(nctNumber);
            if (rowId === undefined) {
                this.log('Ignoring invalid study object from server: NCT ID %s is not valid!', nctNumber);
                return;
            }
            yield db.run('INSERT INTO ctgov_studies (nct_id, study_json, created_at) VALUES (?, ?, ?) ON CONFLICT(nct_id) DO UPDATE SET study_json=excluded.study_json', rowId, JSON.stringify(study), new Date().valueOf());
        });
    }
    /**
     * Loads a ClinicalStudy from an extracted dataset. This will never download a copy, this will only ever return from
     * within the cache.
     * @param nctNumber the NCT number
     * @returns a Promise that resolves to either the parsed ClinicalStudy or to null if the ClinicalStudy does not exist
     */
    getCachedClinicalStudy(nctNumber) {
        return __awaiter(this, void 0, void 0, function* () {
            const db = this.getDB();
            const nctId = parseNCTNumber(nctNumber);
            if (nctId) {
                const result = yield db.get('SELECT study_json FROM ctgov_studies WHERE nct_id=?', nctId);
                // study_json can be NULL to indicate a cached failure
                if (result && result.study_json) {
                    return JSON.parse(result.study_json);
                }
            }
            return null;
        });
    }
    /**
     * Creates and initializes a new service for retrieving data from http://clinicaltrials.gov/.
     *
     * This is essentially the same as constructing the object and then calling init() on it.
     *
     * Note the same behavior applies when `create()` is invoked with a database: the database object is nulled out when
     * {@link #destroy} is called, but it isn't closed. When invoked with a database object, it is the caller's
     * responsibility to close it.
     *
     * @param db the path to the database file or the database to use directly
     * @param options additional options that can be set to further configure the trial service
     * @returns a Promise that resolves when the service is ready
     */
    static create(db, options) {
        return __awaiter(this, void 0, void 0, function* () {
            const result = new ClinicalTrialsGovService(db, options);
            yield result.init();
            return result;
        });
    }
}
exports.ClinicalTrialsGovService = ClinicalTrialsGovService;
/**
 * Alias of ClinicalTrialsGovService.create.
 */
exports.createClinicalTrialsGovService = ClinicalTrialsGovService.create;
